/*-
 * Copyright (c) 2009 OKA Shinichiro.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <inttypes.h>

#include "B25sink.hpp"


const int
B25sink::round = 4;

const int
B25sink::strip = 0;

const int
B25sink::emm = 0;


void
B25sink::put(const uint8_t *buf, int32_t len)
{
	ARIB_STD_B25_BUFFER sbuf;
	sbuf.data = const_cast<uint8_t*>(buf);
	sbuf.size = len;

	int code;
	code = _b25->put(_b25, &sbuf);
	if (code < 0)
		throw "error - failed on ARIB_STD_B25::put()";

	get();
}

void
B25sink::get()
{
	ARIB_STD_B25_BUFFER dbuf;
	int code = _b25->get(_b25, &dbuf);
	if (code < 0)
		throw "error - failed on ARIB_STD_B25::get()";

	if (dbuf.size > 0)
		_sink->put(dbuf.data, dbuf.size);
}

void
B25sink::flush()
{
	int code = _b25->flush(_b25);
	if (code < 0)
		throw "error - failed on ARIB_STD_B25::flush()";

	get();

	_sink->flush();
}

void
B25sink::setChain(Datasink* sink)
{
	_sink = sink;
}

B25sink::B25sink() : _sink(NULL), _b25(NULL), _bcas(NULL)
{
	_b25 = create_arib_std_b25();
	if (_b25 == NULL)
		throw "error - failed on create_arib_std_b25()";

	int code;
	code = _b25->set_multi2_round(_b25, B25sink::round);
	if (code < 0)
		throw "error - failed on ARIB_STD_B25::set_multi2_round()";

	code = _b25->set_strip(_b25, B25sink::strip);
	if (code < 0)
		throw "error - failed on ARIB_STD_B25::set_strip()";

	code = _b25->set_emm_proc(_b25, B25sink::emm);
	if (code < 0)
		throw "error - failed on ARIB_STD_B25::set_emm_proc()";

	_bcas = create_b_cas_card();
	if (_bcas == NULL)
		throw "error - failed on create_b_cas_card()";

	code = _bcas->init(_bcas);
	if (code < 0)
		throw "error - failed on B_CAS_CARD::init()";

	code = _b25->set_b_cas_card(_b25, _bcas);
	if (code < 0)
		throw "error - failed on ARIB_STD_B25::set_b_cas_card()";
}

B25sink::~B25sink()
{
	if (_b25 != NULL)
		_b25->release(_b25);
	if (_bcas != NULL)
		_bcas->release(_bcas);
}
