/*-
 * Copyright (c) 2009 OKA Shinichiro.
 * All rights reserved.
 * Copyright (c) 2008 FUKAUMI Naoki.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <cstdio>
#include <cstdlib>
#include <cstring>

#include <inttypes.h>
#include <openssl/des.h>

#include "Decrypt.hpp"

#define TS_PKT_LEN 188

/* XXX: these are intermediate data, may not work on another architecture */
static const uint32_t ks[2][32] = {
      { 0xb0d03408, 0xcd06ce01, 0x08cc18b8, 0xc2424349,
	0xb448204c, 0x0ac68943, 0x602490d0, 0xcb0fc101,
	0x1c0c0094, 0x48854aca, 0x546cf454, 0xc440c041,
	0x343014e4, 0x8a898e0b, 0x4020b880, 0xccc18cc4,
	0xf0646030, 0x860b864e, 0x50b028e0, 0x03c8c088,
	0xa82c448c, 0x044bc0cc, 0x30d450e4, 0x054205c3,
	0xf8385c60, 0x0e054844, 0x00f8807c, 0x41034c0d,
	0x8c8c9c04, 0x0e430305, 0x6c48c8bc, 0x0984840c, },

      { 0x000044c0, 0xc14089c0, 0x8024200c, 0x82090285,
	0x10807400, 0x098a8180, 0x882444a0, 0x440c0c80,
	0x1450c830, 0x010044c2, 0x4820486c, 0x86020204,
	0x00d03028, 0x00838108, 0x80040080, 0x064e4904,
	0x2428386c, 0x0080050c, 0x00801808, 0x804d8901,
	0xa0088098, 0x0c860982, 0x4010c410, 0x41040341,
	0x0c082070, 0x0804c60e, 0x40c02894, 0xc0498001,
	0xa4001088, 0x0a8a80c8, 0x14d8a870, 0x00094200, },
};

void
Decryptor::decrypt(const uint8_t* buf, int32_t len)
{
	// ɤ߹Ĥȹ碌Ƥ188byte­ʤ
	if (TS_PKT_LEN > _pkt_bpos + len) {
		memcpy(&_pkt.b[4 + _pkt_bpos], buf, len);
		_pkt_bpos += len;
		return;
	}

	// ɤ߹Ĥ꤬
	if (_pkt_bpos > 0) {
		int chunk = (TS_PKT_LEN - _pkt_bpos);
		memcpy(&_pkt.b[4 + _pkt_bpos], buf, chunk);
		decode(&(_pkt.b[4]));

		buf += chunk;
		len -= chunk;
	}

	// SYNC byteǻϤޤ188byteȤˡǥɤ
	while (len >= TS_PKT_LEN) {
		if (*buf != 0x47) {
#if 0
			fprintf(stderr, "?"); // seek to SYNC byte
#endif
			++buf;
			--len;
			continue;
		}

		decode(const_cast<uint8_t*>(buf));

		buf += TS_PKT_LEN;
		len -= TS_PKT_LEN;
	}

	// Ǹ188byte̤ΥǡϿ(SYNC byteޤǤskip)
	while (len > 0) {
		if (*buf != 0x47) {
#if 0
			fprintf(stderr, "!"); // seek to SYNC byte
#endif
			++buf;
			--len;
			continue;
		}

		memcpy(&_pkt.b[4], buf, len);
		break;
	}
	_pkt_bpos = len;
}

// TSѥåȤ
void
Decryptor::decode(uint8_t* buf)
{
	uint8_t wbuf[TS_PKT_LEN];
	int i;

	// Ƭ4byteϤΤޤ
	for (i = 0; i < 4; ++i)
		wbuf[i] = buf[i];

	uint64_t* src = (uint64_t*)&buf[4];
	DES_cblock* dst = (DES_cblock*)&wbuf[4];

	// Ĥ188-4byteϡXOR+DES
	for (i = 0; i < 16; ++i) {
		src[i] ^= xorblk.l[0];
		DES_ecb_encrypt((DES_cblock *)&src[i], &dst[i], &dks[0], DES_DECRYPT);
	}
	for (; i < 23; ++i) {
		src[i] ^= xorblk.l[0];
		DES_ecb_encrypt((DES_cblock *)&src[i], &dst[i], &dks[1], DES_DECRYPT);
	}

	_sink->put(wbuf, TS_PKT_LEN);
}

// ǡΥեå
void
Decryptor::flush()
{
	// 󥹥󥹤˻äƤ롢188byte̤Υǡ˴
	_sink->flush();
}

Decryptor::Decryptor(Datasink* sink) : _sink(sink), _pkt_bpos(0)
{
	xorblk.b[0] = 0x00;
	xorblk.b[1] = 0x00;
	xorblk.b[2] = 0xb1;
	xorblk.b[3] = 0xf2;
	xorblk.b[4] = 0x00;
	xorblk.b[5] = 0x04;
	xorblk.b[6] = 0xf2;
	xorblk.b[7] = 0x04;

	memcpy(dks[0].ks->deslong, ks[0], sizeof(ks[0]));
	memcpy(dks[1].ks->deslong, ks[1], sizeof(ks[1]));
}
