/*-
 * Copyright (c) 2009 OKA Shinichiro.
 * All rights reserved.
 * Copyright (c) 2008 FUKAUMI Naoki.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <sstream>

#include <aio.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <err.h>
#include <fcntl.h>
#include <inttypes.h>
#include <limits.h>
#include <unistd.h>
#include <signal.h>
#include <err.h>

#include <libusb20.h>
#include <libusb20_desc.h>

#include <dev/usb/usb.h>
#include <dev/usb/usb_ioctl.h>

#include "Hdus.hpp"

#include "Decrypt.hpp"
#include "Filesink.hpp"
#include "Bufsink.hpp"
#ifdef B25
# include "B25sink.hpp"
#endif

#include "asio.hpp"

// as11loader¹ԸΥǥХ
static struct {
	uint16_t vendor;
	uint16_t product;
} device_ids[] = {
	// SKNET HDUSF
	{ vendor : 0x3275,
	  product : 0x6051},

	// Quixun QRS-UT100B
	{ vendor : 0x3275,
	  product : 0x7020},

	{ vendor : 0,
	  product : 0}
};

int
Hdus::send_ctl(uint8_t *data, int req, int val, int idx, int len)
{
	if (_ep0 < 0)
		err(EXIT_FAILURE, "invalid status (no available device)");

	struct usb2_ctl_request ucr;

	memset(&ucr, 0, sizeof(ucr));
	ucr.ucr_request.bmRequestType = UT_READ_VENDOR_DEVICE;
	ucr.ucr_request.bRequest = req;
	USETW(ucr.ucr_request.wValue, val);
	USETW(ucr.ucr_request.wIndex, idx);
	USETW(ucr.ucr_request.wLength, len);
	ucr.ucr_data = data;

	int r = ioctl(_ep0, USB_DO_REQUEST, &ucr);

	if (req == 0x08)
		usleep(10000);

	return r;
}

// ProductId פǥХȤ
// BUSYʤ鼡õ
// ĤʤСthrow
void
Hdus::open() throw (const char*)
{
	struct libusb20_backend *pbe = libusb20_be_alloc_default();
	if (pbe == NULL)
		err(EXIT_FAILURE, "could not initialize usb backend");

	struct libusb20_device *pdev = NULL;
	while ((pdev = libusb20_be_device_foreach(pbe, pdev))) {
		if (libusb20_dev_open(pdev, 0))
			continue; // XXX

		if (is_hdus(pdev))
			open_hdus(pdev);

		if (libusb20_dev_close(pdev)) {
			err(EXIT_FAILURE, "could not close device");
		}

		if (_ep0 >= 0)
			break;
	}

	libusb20_be_free(pbe);

	if (_ep0 < 0)
		throw "could not open device";
}

bool
Hdus::is_hdus(struct libusb20_device *pdev)
{
	struct LIBUSB20_DEVICE_DESC_DECODED *ddesc =
	    libusb20_dev_get_device_desc(pdev);

	int i = 0;
	while (device_ids[i].vendor) {
		if (ddesc->idVendor == device_ids[i].vendor &&
		    ddesc->idProduct == device_ids[i].product)
			return true;
		++i;
	}

	return false;
}

void
Hdus::open_hdus(struct libusb20_device *pdev)
{
	int bus = libusb20_dev_get_bus_number(pdev);
	int addr = libusb20_dev_get_address(pdev);

	int fd;

	// ep0ʣΥץopenǤ뤬
	// ep1ϣĤΥץ餷openǤʤΤǡ
	// ¾Ͽץ¹ԤƤʤǧ

	// ep1 for read
	std::ostringstream ep1_path;
	ep1_path << "/dev/usb/" << bus << "." << addr << ".1";
	fd = ::open(ep1_path.str().c_str(), O_RDONLY|O_EXLOCK|O_DIRECT);
	if (fd <= 0)
		return; // fail
	_ep1 = fd;

	// ep0 for control
	std::ostringstream ep0_path;
	ep0_path << "/dev/usb/" << bus << "." << addr << ".0";
	fd = ::open(ep0_path.str().c_str(), O_WRONLY);
	if (fd <= 0) {
		::close(_ep1);
		_ep1 = -1;
		return; // fail
	}

	_ep0 = fd;
}

void
Hdus::close()
{
	if (_ep0 >= 0) {
		::close(_ep0);
		_ep0 = -1;
	}
	if (_ep1 >= 0) {
		::close(_ep1);
		_ep1 = -1;
	}
}

void
Hdus::init()
{
	uint8_t data[0x200];

	send_ctl(data, 0x0c, 0x0000, 0x0000, 0x003a);
	send_ctl(data, 0x17, 0x0002, 0x0000, 0x0003);
	send_ctl(data, 0x08, 0xfb07, 0x0000, 0x0001);

	/* HDUS: Tuner off */
	send_ctl(data, 0x08, 0x0100, 0x0000, 0x0001);
	send_ctl(data, 0x08, 0xdd0c, 0x0000, 0x0001);

	/* HDUS: Tuner on */
	send_ctl(data, 0x08, 0xd500, 0x0000, 0x0001);
	send_ctl(data, 0x08, 0x0800, 0x0000, 0x0001);
	send_ctl(data, 0x08, 0x4040, 0x0000, 0x0001);
	send_ctl(data, 0x08, 0x8080, 0x0000, 0x0001);
	send_ctl(data, 0x08, 0x0404, 0x0000, 0x0001);

	/* HDUS: Tuner init */
	send_ctl(data, 0x03, 0x0120, 0x0000, 0x0002);
	send_ctl(data, 0x03, 0x0220, 0x0000, 0x0002);
	send_ctl(data, 0x03, 0x0320, 0x0000, 0x0002);
	send_ctl(data, 0x03, 0x0420, 0x0000, 0x0002);
	send_ctl(data, 0x03, 0x0520, 0x0000, 0x0002);
	send_ctl(data, 0x03, 0x0620, 0x0000, 0x0002);
	send_ctl(data, 0x03, 0x0720, 0x0000, 0x0002);
	send_ctl(data, 0x03, 0x0820, 0x0000, 0x0002);
	send_ctl(data, 0x03, 0x0c20, 0x0000, 0x0002);
	send_ctl(data, 0x03, 0x1120, 0x0021, 0x0002);
	send_ctl(data, 0x03, 0x1220, 0x000b, 0x0002);
	send_ctl(data, 0x03, 0x1320, 0x002c, 0x0002);
	send_ctl(data, 0x03, 0x1420, 0x000a, 0x0002);
	send_ctl(data, 0x03, 0x1520, 0x0040, 0x0002);
	send_ctl(data, 0x03, 0x1620, 0x000a, 0x0002);
	send_ctl(data, 0x03, 0x1720, 0x0070, 0x0002);
	send_ctl(data, 0x03, 0x1820, 0x0031, 0x0002);
	send_ctl(data, 0x03, 0x1920, 0x0013, 0x0002);
	send_ctl(data, 0x03, 0x1a20, 0x0031, 0x0002);
	send_ctl(data, 0x03, 0x1b20, 0x0013, 0x0002);
	send_ctl(data, 0x03, 0x1c20, 0x002a, 0x0002);
	send_ctl(data, 0x03, 0x1d20, 0x00aa, 0x0002);
	send_ctl(data, 0x03, 0x1e20, 0x00a2, 0x0002);
	send_ctl(data, 0x03, 0x1f20, 0x0008, 0x0002);
	send_ctl(data, 0x03, 0x2020, 0x0000, 0x0002);
	send_ctl(data, 0x03, 0x2120, 0x00ff, 0x0002);
	send_ctl(data, 0x03, 0x2220, 0x0080, 0x0002);
	send_ctl(data, 0x03, 0x2320, 0x004c, 0x0002);
	send_ctl(data, 0x03, 0x2420, 0x004c, 0x0002);
	send_ctl(data, 0x03, 0x2520, 0x0000, 0x0002);
	send_ctl(data, 0x03, 0x2620, 0x0000, 0x0002);
	send_ctl(data, 0x03, 0x2720, 0x000c, 0x0002);
	send_ctl(data, 0x03, 0x2820, 0x0060, 0x0002);
	send_ctl(data, 0x03, 0x2920, 0x006b, 0x0002);
	send_ctl(data, 0x03, 0x2a20, 0x0040, 0x0002);
	send_ctl(data, 0x03, 0x2b20, 0x0040, 0x0002);
	send_ctl(data, 0x03, 0x2c20, 0x00ff, 0x0002);
	send_ctl(data, 0x03, 0x2d20, 0x0000, 0x0002);
	send_ctl(data, 0x03, 0x2e20, 0x00ff, 0x0002);
	send_ctl(data, 0x03, 0x2f20, 0x0000, 0x0002);
	send_ctl(data, 0x03, 0x3020, 0x0028, 0x0002);
	send_ctl(data, 0x03, 0x3120, 0x000f, 0x0002);
	send_ctl(data, 0x03, 0x3220, 0x00a0, 0x0002);
	send_ctl(data, 0x03, 0x3420, 0x003f, 0x0002);
	send_ctl(data, 0x03, 0x3820, 0x0001, 0x0002);
	send_ctl(data, 0x03, 0x3920, 0x0070, 0x0002);
	send_ctl(data, 0x03, 0x3a20, 0x002d, 0x0002);
	send_ctl(data, 0x03, 0x3b20, 0x0021, 0x0002);
	send_ctl(data, 0x03, 0x3c20, 0x003f, 0x0002);
	send_ctl(data, 0x03, 0x3d20, 0x0010, 0x0002);
	send_ctl(data, 0x03, 0x3e20, 0x0008, 0x0002);
	send_ctl(data, 0x03, 0x3f20, 0x000c, 0x0002);
	send_ctl(data, 0x03, 0x4020, 0x000c, 0x0002);
	send_ctl(data, 0x03, 0x4120, 0x0000, 0x0002);
	send_ctl(data, 0x03, 0x4220, 0x0000, 0x0002);
	send_ctl(data, 0x03, 0x4320, 0x004f, 0x0002);
	send_ctl(data, 0x03, 0x4420, 0x00ff, 0x0002);
	send_ctl(data, 0x03, 0x4620, 0x0020, 0x0002);
	send_ctl(data, 0x03, 0x4720, 0x0000, 0x0002);
	send_ctl(data, 0x03, 0x4820, 0x0090, 0x0002);
	send_ctl(data, 0x03, 0x4920, 0x00e6, 0x0002);
	send_ctl(data, 0x03, 0x4a20, 0x0002, 0x0002);
	send_ctl(data, 0x03, 0x4b20, 0x0054, 0x0002);
	send_ctl(data, 0x03, 0x4c20, 0x0000, 0x0002);
	send_ctl(data, 0x03, 0x5020, 0x0000, 0x0002);
	send_ctl(data, 0x03, 0x5120, 0x0058, 0x0002);
	send_ctl(data, 0x03, 0x5220, 0x0020, 0x0002);
	send_ctl(data, 0x03, 0x5420, 0x0057, 0x0002);
	send_ctl(data, 0x03, 0x5520, 0x00f1, 0x0002);
	send_ctl(data, 0x03, 0x5620, 0x0020, 0x0002);
	send_ctl(data, 0x03, 0x5720, 0x0070, 0x0002);
	send_ctl(data, 0x03, 0x5820, 0x0060, 0x0002);
	send_ctl(data, 0x03, 0x5c20, 0x0050, 0x0002);
	send_ctl(data, 0x03, 0x5d20, 0x0000, 0x0002);
	send_ctl(data, 0x03, 0x5f20, 0x0080, 0x0002);
	send_ctl(data, 0x03, 0x7020, 0x0018, 0x0002);
	send_ctl(data, 0x03, 0x7120, 0x0000, 0x0002);
	send_ctl(data, 0x03, 0x7220, 0x0000, 0x0002);
	send_ctl(data, 0x03, 0x7520, 0x0022, 0x0002);
	send_ctl(data, 0x03, 0x7620, 0x0000, 0x0002);
	send_ctl(data, 0x03, 0x7720, 0x0001, 0x0002);
	send_ctl(data, 0x03, 0x7c20, 0x0000, 0x0002);
	send_ctl(data, 0x03, 0x7d20, 0x0052, 0x0002);
	send_ctl(data, 0x03, 0xba20, 0x0000, 0x0002);
	send_ctl(data, 0x03, 0xbb20, 0x0000, 0x0002);
	send_ctl(data, 0x03, 0xbc20, 0x0000, 0x0002);
	send_ctl(data, 0x03, 0xc220, 0x0000, 0x0002);
	send_ctl(data, 0x03, 0xc720, 0x0000, 0x0002);
	send_ctl(data, 0x03, 0xe420, 0x001a, 0x0002);
	send_ctl(data, 0x03, 0xe920, 0x0008, 0x0002);
	send_ctl(data, 0x03, 0xec20, 0x0000, 0x0002);
	send_ctl(data, 0x03, 0xef20, 0x0000, 0x0002);
}

uint8_t
Hdus::asie5606(int null)
{
	uint8_t data[0x200], tid = 0;
	uint8_t r5 = 0x91, saddr = 0x5a;
	uint32_t val;

	/* ASIE5606 on */
	send_ctl(data, 0x08, ((tid == 0) ? 0x1100 : 0x8000) | 0x00,
	    0x0000, 0x0001);
	send_ctl(data, 0x08, ((tid == 0) ? 0x1100 : 0x8000) | 0xff,
	    0x0000, 0x0001);

	/* ASIE5606 init */
	send_ctl(data, 0x02, 0x0900 | saddr, 0x0000, 0x0002);
	if (data[0] != 1) {
		saddr = 0x4a; /* HDUSF */
		r5 = 0x80; /* DES */
		send_ctl(data, 0x02, 0x0900 | saddr, 0x0000, 0x0002);
		send_ctl(data, 0x23, 0x0000 | saddr, 0x0000, 0x0002);
		send_ctl(data, 0x02, 0x0900 | saddr, 0x0000, 0x0002);
	}
	send_ctl(data, 0x03, 0x0500 | saddr, r5, 0x0002);
	if (r5 == 0x91) {
		/* Set XOR values */
		for (val = 0x1000; val < 0x2000; val += 0x0100)
			send_ctl(data, 0x03, val | saddr,
			    (null == 2) ? 0x00ff : 0x0000, 0x0002);
	} else {
		/* Set key */
		for (val = 0x1000; val < 0x2000; val += 0x0100)
			send_ctl(data, 0x03, val | saddr, 0x0000, 0x0002);
	}

	return saddr;
}

void
Hdus::setfreq(int freq, int xpid_min, int xpid_max)
{
	uint8_t data[0x200], tid = 0;
	uint32_t val;

	/* HDUS: Set frequency */
	val = (freq + 57) * 7 + 1;
	send_ctl(data, 0x0d, 0xfe00, (val & 0xff00) | 0xc0, 0x0004);
	send_ctl(data, 0x0d, ((val & 0xff) << 8) | 0x03,
	    (freq < 205) ? 0x8191 : ((freq < 421) ? 0x8291 : 0x8491), 0x0004);
	send_ctl(data, 0x0d, 0xe306, 0x0000, 0x0002);
	send_ctl(data, 0x0e, 0x0020, 0x0000, 0x0008);
	send_ctl(data, 0x03, 0x0120, 0x0040, 0x0002);

	send_ctl(data, 0x09, 0x0100 | tid, 0x0000, 0x0001);

	/* Set PID filter */
	val = (tid == 0) ? 0x40 : 0xc0;
	send_ctl(data, 0x04, val, 0x0000, 0x0002);
	send_ctl(data, 0x05, ((data[1] | 0x0b) << 8) | val,
	    0x0000, 0x0002);
	send_ctl(data, 0x05, (xpid_min & 0xff00) | (val + 1),
	    (xpid_min & 0xff), 0x0003);
	send_ctl(data, 0x05, (xpid_max & 0xff00) | (val + 3),
	    (xpid_max & 0xff), 0x0003);
}

void
Hdus::start()
{
	uint8_t data[0x200], tid = 0;

	/* Start */
	send_ctl(data, 0x06, tid, 0x0000, 0x0001);
}

void
Hdus::stop(uint8_t saddr)
{
	uint8_t data[0x200], tid = 0;

	/* Stop */
	send_ctl(data, 0x07, tid, 0x0000, 0x0001);

	send_ctl(data, 0x03, 0x8000 | saddr, 0x0000, 0x0002);
	send_ctl(data, 0x03, 0x0400 | saddr, 0x0000, 0x0002);
	send_ctl(data, 0x03, 0x0000 | saddr, 0x0000, 0x0002);
	if (saddr == 0x4a)
		send_ctl(data, 0x21, 0x0000 | saddr, 0x0000, 0x0001);

	/* HDUS: Tuner off */
	send_ctl(data, 0x08, 0x0100, 0x0000, 0x0001);
	send_ctl(data, 0x08, 0xdd0c, 0x0000, 0x0001);
}

void
Hdus::record(int duration, const char* outfile, int w_b25)
{
	// hdus > decrypt > buf > b25 > fileout
	// ޤ
	// hdus > decrypt > buf > fileout

	Filesink ofile(outfile, O_WRONLY|O_CREAT|O_TRUNC, 0666);
	Bufsink buff;
#ifdef B25
	B25sink b25dec;
	if (w_b25) {
		b25dec.setChain(&ofile);
		buff.setChain(&b25dec);
	} else {
		buff.setChain(&ofile);
	}
#else
	buff.setChain(&ofile);
#endif
	Decryptor decrypt(&buff);

	// Ͽ襹
	asio_start(_ep1, duration);
	try {
		while (struct ringbuf* buf = asio_getbuf()) {
			decrypt.decrypt(buf->data, buf->len);
		}
	} catch (const char* e) {
		asio_stop();
		fprintf(stderr, "%s\n", e);
	}

	decrypt.flush();
}
